################################################################################
#                                 Sarimax                                      #
#                                                                              #
# This work by skforecast team is licensed under the BSD 3-Clause License.     #
################################################################################
# coding=utf-8

from typing import Union, Optional
import warnings
import logging
import inspect
import numpy as np
import pandas as pd
from sklearn.base import clone, BaseEstimator, RegressorMixin
from sklearn.exceptions import NotFittedError
from statsmodels.tsa.statespace.sarimax import SARIMAX
import skforecast

logging.basicConfig(
    format = '%(name)-10s %(levelname)-5s %(message)s', 
    level  = logging.INFO,
)


def _check_fitted(func):
    """
    This decorator checks if the model is fitted before using the desired method.

    Parameters
    ----------
    func : Callable
        Function to wrap.
    
    Returns
    -------
    wrapper : wrapper
        Function wrapped.

    """

    def wrapper(self, *args, **kwargs):

        if not self.fitted:
            raise NotFittedError(
                ("Sarimax instance is not fitted yet. Call `fit` with "
                 "appropriate arguments before using this method.")
            )
        
        result = func(self, *args, **kwargs)
        
        return result
    
    return wrapper


class Sarimax(BaseEstimator, RegressorMixin):
    """
    A universal scikit-learn style wrapper for statsmodels SARIMAX.

    This class wraps the statsmodels.tsa.statespace.sarimax.SARIMAX model to 
    follow the scikit-learn style. The following docstring is based on the 
    statmodels documentation and it is highly recommended to visit their site 
    for the best level of detail.

    https://www.statsmodels.org/stable/generated/statsmodels.tsa.statespace.sarimax.SARIMAX.html
    https://www.statsmodels.org/dev/generated/statsmodels.tsa.statespace.sarimax.SARIMAXResults.html

    Parameters
    ----------
    order : tuple, default `(1, 0, 0)`
        The (p,d,q) order of the model for the number of AR parameters, differences, 
        and MA parameters. 
        
        - `d` must be an integer indicating the integration order of the process.
        - `p` and `q` may either be an integers indicating the AR and MA orders 
        (so that all lags up to those orders are included) or else iterables 
        giving specific AR and / or MA lags to include.
    seasonal_order : tuple, default `(0, 0, 0, 0)`
        The (P,D,Q,s) order of the seasonal component of the model for the AR 
        parameters, differences, MA parameters, and periodicity. 
        
        - `D` must be an integer indicating the integration order of the process.
        - `P` and `Q` may either be an integers indicating the AR and MA orders 
        (so that all lags up to those orders are included) or else iterables 
        giving specific AR and / or MA lags to include. 
        - `s` is an integer giving the periodicity (number of periods in season),
        often it is 4 for quarterly data or 12 for monthly data.
    trend : str, default `None`
        Parameter controlling the deterministic trend polynomial `A(t)`.
            
        - `'c'` indicates a constant (i.e. a degree zero component of the 
        trend polynomial).
        - `'t'` indicates a linear trend with time.
        - `'ct'` indicates both, `'c'` and `'t'`. 
        - Can also be specified as an iterable defining the non-zero polynomial 
        exponents to include, in increasing order. For example, `[1,1,0,1]` 
        denotes `a + b*t + ct^3`.
    measurement_error : bool, default `False`
        Whether or not to assume the endogenous observations `y` were measured 
        with error.
    time_varying_regression : bool, default `False`
        Used when an explanatory variables, `exog`, are provided to select whether 
        or not coefficients on the exogenous regressors are allowed to vary over time.
    mle_regression : bool, default `True`
        Whether or not to use estimate the regression coefficients for the
        exogenous variables as part of maximum likelihood estimation or through
        the Kalman filter (i.e. recursive least squares). If 
        `time_varying_regression` is `True`, this must be set to `False`.
    simple_differencing : bool, default `False`
        Whether or not to use partially conditional maximum likelihood
        estimation. 
        
        - If `True`, differencing is performed prior to estimation, which 
        discards the first `s*D + d` initial rows but results in a smaller 
        state-space formulation. 
        - If `False`, the full SARIMAX model is put in state-space form so 
        that all datapoints can be used in estimation.
    enforce_stationarity : bool, default `True`
        Whether or not to transform the AR parameters to enforce stationarity
        in the autoregressive component of the model.
    enforce_invertibility : bool, default `True`
        Whether or not to transform the MA parameters to enforce invertibility
        in the moving average component of the model.
    hamilton_representation : bool, default `False`
        Whether or not to use the Hamilton representation of an ARMA process
        (if `True`) or the Harvey representation (if `False`).
    concentrate_scale : bool, default `False`
        Whether or not to concentrate the scale (variance of the error term)
        out of the likelihood. This reduces the number of parameters estimated
        by maximum likelihood by one, but standard errors will then not
        be available for the scale parameter.
    trend_offset : int, default `1`
        The offset at which to start time trend values. Default is 1, so that
        if `trend='t'` the trend is equal to 1, 2, ..., nobs. Typically is only
        set when the model created by extending a previous dataset.
    use_exact_diffuse : bool, default `False`
        Whether or not to use exact diffuse initialization for non-stationary
        states. Default is `False` (in which case approximate diffuse
        initialization is used).
    method : str, default `'lbfgs'`
        The method determines which solver from scipy.optimize is used, and it 
        can be chosen from among the following strings:

        - `'newton'` for Newton-Raphson
        - `'nm'` for Nelder-Mead
        - `'bfgs'` for Broyden-Fletcher-Goldfarb-Shanno (BFGS)
        - `'lbfgs'` for limited-memory BFGS with optional box constraints
        - `'powell'` for modified Powell`s method
        - `'cg'` for conjugate gradient
        - `'ncg'` for Newton-conjugate gradient
        - `'basinhopping'` for global basin-hopping solver
    maxiter : int, default `50`
        The maximum number of iterations to perform.
    start_params : numpy ndarray, default `None`
        Initial guess of the solution for the loglikelihood maximization. 
        If `None`, the default is given by regressor.start_params.
    disp : bool, default `False`
        Set to `True` to print convergence messages.
    sm_init_kwargs : dict, default `{}`
        Additional keyword arguments to pass to the statsmodels SARIMAX model 
        when it is initialized.
    sm_fit_kwargs : dict, default `{}` 
        Additional keyword arguments to pass to the `fit` method of the
        statsmodels SARIMAX model. The statsmodels SARIMAX.fit parameters 
        `method`, `max_iter`, `start_params` and `disp` have been moved to the 
        initialization of this model and will have priority over those provided 
        by the user using via `sm_fit_kwargs`.
    sm_predict_kwargs : dict, default `{}`
        Additional keyword arguments to pass to the `get_forecast` method of the
        statsmodels SARIMAXResults object.

    Attributes
    ----------
    order : tuple
        The (p,d,q) order of the model for the number of AR parameters, differences, 
        and MA parameters.
    seasonal_order : tuple
        The (P,D,Q,s) order of the seasonal component of the model for the AR 
        parameters, differences, MA parameters, and periodicity.
    trend : str
        Deterministic trend polynomial `A(t)`.
    measurement_error : bool
        Whether or not to assume the endogenous observations `y` were measured 
        with error.
    time_varying_regression : bool
        Used when an explanatory variables, `exog`, are provided to select whether 
        or not coefficients on the exogenous regressors are allowed to vary over time.
    mle_regression : bool
        Whether or not to use estimate the regression coefficients for the
        exogenous variables as part of maximum likelihood estimation or through
        the Kalman filter (i.e. recursive least squares). If 
        `time_varying_regression` is `True`, this must be set to `False`.
    simple_differencing : bool
        Whether or not to use partially conditional maximum likelihood
        estimation.
    enforce_stationarity : bool
        Whether or not to transform the AR parameters to enforce stationarity
        in the autoregressive component of the model.
    enforce_invertibility : bool
        Whether or not to transform the MA parameters to enforce invertibility
        in the moving average component of the model.
    hamilton_representation : bool
        Whether or not to use the Hamilton representation of an ARMA process
        (if `True`) or the Harvey representation (if `False`).
    concentrate_scale : bool
        Whether or not to concentrate the scale (variance of the error term)
        out of the likelihood. This reduces the number of parameters estimated
        by maximum likelihood by one, but standard errors will then not
        be available for the scale parameter.
    trend_offset : int
        The offset at which to start time trend values.
    use_exact_diffuse : bool
        Whether or not to use exact diffuse initialization for non-stationary
        states.
    method : str
        The method determines which solver from scipy.optimize is used.
    maxiter : int
        The maximum number of iterations to perform.
    start_params : numpy ndarray
        Initial guess of the solution for the loglikelihood maximization.
    disp : bool
        Set to `True` to print convergence messages.
    sm_init_kwargs : dict
        Additional keyword arguments to pass to the statsmodels SARIMAX model 
        when it is initialized.
    sm_fit_kwargs : dict
        Additional keyword arguments to pass to the `fit` method of the
        statsmodels SARIMAX model.
    sm_predict_kwargs : dict
        Additional keyword arguments to pass to the `get_forecast` method of the
        statsmodels SARIMAXResults object.
    _sarimax_params : dict
        Parameters of this model that can be set with the `set_params` method.
    output_type : str
        Format of the object returned by the predict method. This is set 
        automatically according to the type of `y` used in the fit method to 
        train the model, `'numpy'` or `'pandas'`.
    sarimax : object
        The statsmodels.tsa.statespace.sarimax.SARIMAX object created.
    fitted : bool
        Tag to identify if the regressor has been fitted (trained).
    sarimax_res : object
        The resulting statsmodels.tsa.statespace.sarimax.SARIMAXResults object 
        created by statsmodels after fitting the SARIMAX model.
    training_index : pandas Index
        Index of the training series as long as it is a pandas Series or Dataframe.

    """

    def __init__(
        self,
        order: tuple = (1, 0, 0),
        seasonal_order: tuple = (0, 0, 0, 0),
        trend: str = None,
        measurement_error: bool = False,
        time_varying_regression: bool = False,
        mle_regression: bool = True,
        simple_differencing: bool = False,
        enforce_stationarity: bool = True,
        enforce_invertibility: bool = True,
        hamilton_representation: bool = False,
        concentrate_scale: bool = False,
        trend_offset: int = 1,
        use_exact_diffuse: bool = False,
        dates = None,
        freq = None,
        missing = 'none',
        validate_specification: bool = True,
        method: str = 'lbfgs',
        maxiter: int = 50,
        start_params: np.ndarray = None,
        disp: bool = False,
        sm_init_kwargs: dict = {},
        sm_fit_kwargs: dict = {},
        sm_predict_kwargs: dict = {}
    ) -> None:

        self.order                   = order
        self.seasonal_order          = seasonal_order
        self.trend                   = trend
        self.measurement_error       = measurement_error
        self.time_varying_regression = time_varying_regression
        self.mle_regression          = mle_regression
        self.simple_differencing     = simple_differencing
        self.enforce_stationarity    = enforce_stationarity
        self.enforce_invertibility   = enforce_invertibility
        self.hamilton_representation = hamilton_representation
        self.concentrate_scale       = concentrate_scale
        self.trend_offset            = trend_offset
        self.use_exact_diffuse       = use_exact_diffuse
        self.dates                   = dates
        self.freq                    = freq
        self.missing                 = missing
        self.validate_specification  = validate_specification
        self.method                  = method
        self.maxiter                 = maxiter
        self.start_params            = start_params
        self.disp                    = disp

        # Create the dictionaries with the additional statsmodels parameters to be  
        # used during the init, fit and predict methods. Note that the statsmodels 
        # SARIMAX.fit parameters `method`, `max_iter`, `start_params` and `disp` 
        # have been moved to the initialization of this model and will have 
        # priority over those provided by the user using via `sm_fit_kwargs`.
        self.sm_init_kwargs    = sm_init_kwargs
        self.sm_fit_kwargs     = sm_fit_kwargs
        self.sm_predict_kwargs = sm_predict_kwargs

        # Params that can be set with the `set_params` method
        _, _, _, _sarimax_params = inspect.getargvalues(inspect.currentframe())
        _sarimax_params.pop("self")
        self._sarimax_params = _sarimax_params

        self._consolidate_kwargs()

        # Create Results Attributes 
        self.output_type    = None
        self.sarimax        = None
        self.fitted         = False
        self.sarimax_res    = None
        self.training_index = None


    def __repr__(
        self
    ) -> str:
        """
        Information displayed when a Sarimax object is printed.
        """

        p, d, q = self.order
        P, D, Q, m = self.seasonal_order
        
        return f"Sarimax({p},{d},{q})({P},{D},{Q})[{m}]"


    def _consolidate_kwargs(
        self
    ) -> None:
        """
        Create the dictionaries to be used during the init, fit, and predict methods.
        Note that the parameters in this model's initialization take precedence 
        over those provided by the user using via the statsmodels kwargs dicts.

        Parameters
        ----------
        self
        
        Returns
        -------
        None
        
        """
        
        # statsmodels.tsa.statespace.SARIMAX parameters
        _init_kwargs = self.sm_init_kwargs.copy()
        _init_kwargs.update({
           'order': self.order,
           'seasonal_order': self.seasonal_order,
           'trend': self.trend,
           'measurement_error': self.measurement_error,
           'time_varying_regression': self.time_varying_regression,
           'mle_regression': self.mle_regression,
           'simple_differencing': self.simple_differencing,
           'enforce_stationarity': self.enforce_stationarity,
           'enforce_invertibility': self.enforce_invertibility,
           'hamilton_representation': self.hamilton_representation,
           'concentrate_scale': self.concentrate_scale,
           'trend_offset': self.trend_offset,
           'use_exact_diffuse': self.use_exact_diffuse,
           'dates': self.dates,
           'freq': self.freq,
           'missing': self.missing,
           'validate_specification': self.validate_specification
        })
        self._init_kwargs = _init_kwargs

        # statsmodels.tsa.statespace.SARIMAX.fit parameters
        _fit_kwargs = self.sm_fit_kwargs.copy()
        _fit_kwargs.update({
           'method': self.method,
           'maxiter': self.maxiter,
           'start_params': self.start_params,
           'disp': self.disp,
        })        
        self._fit_kwargs = _fit_kwargs

        # statsmodels.tsa.statespace.SARIMAXResults.get_forecast parameters
        self._predict_kwargs = self.sm_predict_kwargs.copy()


    def _create_sarimax(
        self,
        endog: Union[np.ndarray, pd.Series, pd.DataFrame],
        exog: Optional[Union[np.ndarray, pd.Series, pd.DataFrame]]=None
    ) -> None:
        """
        A helper method to create a new statsmodel SARIMAX model.

        Additional keyword arguments to pass to the statsmodels SARIMAX model 
        when it is initialized can be added with the `init_kwargs` argument 
        when initializing the model.

        Parameters
        ----------
        endog : numpy ndarray, pandas Series, pandas DataFrame
            The endogenous variable.
        exog : numpy ndarray, pandas Series, pandas DataFrame, default `None`
            The exogenous variables.
        
        Returns
        -------
        None

        """

        self.sarimax = SARIMAX(endog=endog, exog=exog, **self._init_kwargs)


    def fit(
        self,
        y: Union[np.ndarray, pd.Series, pd.DataFrame],
        exog: Optional[Union[np.ndarray, pd.Series, pd.DataFrame]]=None
    ) -> None:
        """
        Fit the model to the data.

        Additional keyword arguments to pass to the `fit` method of the
        statsmodels SARIMAX model can be added with the `fit_kwargs` argument 
        when initializing the model.

        Parameters
        ----------
        y : numpy ndarray, pandas Series, pandas DataFrame
            Training time series.
        exog : numpy ndarray, pandas Series, pandas DataFrame, default `None`
            Exogenous variable/s included as predictor/s. Must have the same
            number of observations as `y` and their indexes must be aligned so
            that y[i] is regressed on exog[i].

        Returns
        -------
        None

        """

        # Reset values in case the model has already been fitted.
        self.output_type    = None
        self.sarimax_res    = None
        self.fitted         = False
        self.training_index = None

        self.output_type = 'numpy' if isinstance(y, np.ndarray) else 'pandas'
        
        self._create_sarimax(endog=y, exog=exog)
        self.sarimax_res = self.sarimax.fit(**self._fit_kwargs)
        self.fitted = True
        
        if self.output_type == 'pandas':
            self.training_index = y.index


    @_check_fitted
    def predict(
        self,
        steps: int,
        exog: Optional[Union[np.ndarray, pd.Series, pd.DataFrame]]=None, 
        return_conf_int: bool=False,
        alpha: float=0.05
    ) -> Union[np.ndarray, pd.DataFrame]:
        """
        Forecast future values and, if desired, their confidence intervals.

        Generate predictions (forecasts) n steps in the future with confidence
        intervals. Note that if exogenous variables were used in the model fit, 
        they will be expected for the predict procedure and will fail otherwise.

        Additional keyword arguments to pass to the `get_forecast` method of the
        statsmodels SARIMAX model can be added with the `predict_kwargs` argument 
        when initializing the model.

        Parameters
        ----------
        steps : int
            Number of future steps predicted.
        exog : numpy ndarray, pandas Series, pandas DataFrame, default `None`
            Value of the exogenous variable/s for the next steps. The number of 
            observations needed is the number of steps to predict. 
        return_conf_int : bool, default `False`
            Whether to get the confidence intervals of the forecasts.
        alpha : float, default `0.05`
            The confidence intervals for the forecasts are (1 - alpha) %.

        Returns
        -------
        predictions : numpy ndarray, pandas DataFrame
            Values predicted by the forecaster and their estimated interval. The 
            output type is the same as the type of `y` used in the fit method.

            - pred: predictions.
            - lower_bound: lower bound of the interval. (if `return_conf_int`)
            - upper_bound: upper bound of the interval. (if `return_conf_int`)

        """

        # This is done because statsmodels doesn't allow `exog` length greater than
        # the number of steps
        if exog is not None and len(exog) > steps:
            warnings.warn(
                (f"When predicting using exogenous variables, the `exog` parameter "
                 f"must have the same length as the number of predicted steps. Since "
                 f"len(exog) > steps, only the first {steps} observations are used.")
            )
            exog = exog[:steps]

        predictions = self.sarimax_res.get_forecast(
                          steps = steps,
                          exog  = exog,
                          **self._predict_kwargs
                      )
        
        if not return_conf_int:
            predictions = predictions.predicted_mean
            if self.output_type == 'pandas':
                predictions = predictions.rename("pred").to_frame()
        else:
            if self.output_type == 'numpy':
                predictions = np.column_stack(
                                  [predictions.predicted_mean,
                                   predictions.conf_int(alpha=alpha)]
                              )
            else:
                predictions = pd.concat((
                                  predictions.predicted_mean,
                                  predictions.conf_int(alpha=alpha)),
                                  axis = 1
                              )
                predictions.columns = ['pred', 'lower_bound', 'upper_bound']

        return predictions


    @_check_fitted
    def append(
        self,
        y: Union[np.ndarray, pd.Series, pd.DataFrame],
        exog: Optional[Union[np.ndarray, pd.Series, pd.DataFrame]]=None,
        refit: bool=False,
        copy_initialization: bool=False,
        **kwargs
    ) -> None:
        """
        Recreate the results object with new data appended to the original data.

        Creates a new result object applied to a dataset that is created by 
        appending new data to the end of the model's original data. The new 
        results can then be used for analysis or forecasting.

        Parameters
        ----------
        y : numpy ndarray, pandas Series, pandas DataFrame
            New observations from the modeled time-series process.
        exog : numpy ndarray, pandas Series, pandas DataFrame, default `None`
            New observations of exogenous regressors, if applicable. Must have 
            the same number of observations as `y` and their indexes must be 
            aligned so that y[i] is regressed on exog[i].
        refit : bool, default `False`
            Whether to re-fit the parameters, based on the combined dataset.
        copy_initialization : bool, default `False`
            Whether or not to copy the initialization from the current results 
            set to the new model. 
        **kwargs
            Keyword arguments may be used to modify model specification arguments 
            when created the new model object.

        Returns
        -------
        None

        Notes
        -----
        The `y` and `exog` arguments to this method must be formatted in the same 
        way (e.g. Pandas Series versus Numpy array) as were the `y` and `exog` 
        arrays passed to the original model.

        The `y` argument to this method should consist of new observations that 
        occurred directly after the last element of `y`. For any other kind of 
        dataset, see the apply method.

        This method will apply filtering to all of the original data as well as 
        to the new data. To apply filtering only to the new data (which can be 
        much faster if the original dataset is large), see the extend method.

        https://www.statsmodels.org/dev/generated/statsmodels.tsa.statespace.mlemodel.MLEResults.append.html#statsmodels.tsa.statespace.mlemodel.MLEResults.append

        """

        fit_kwargs = self._fit_kwargs if refit else None

        self.sarimax_res = self.sarimax_res.append(
                               endog               = y,
                               exog                = exog,
                               refit               = refit,
                               copy_initialization = copy_initialization,
                               fit_kwargs          = fit_kwargs,
                               **kwargs
                           )


    @_check_fitted
    def apply(
        self,
        y: Union[np.ndarray, pd.Series, pd.DataFrame],
        exog: Optional[Union[np.ndarray, pd.Series, pd.DataFrame]]=None,
        refit: bool=False,
        copy_initialization: bool=False,
        **kwargs
    ) -> None:
        """
        Apply the fitted parameters to new data unrelated to the original data.

        Creates a new result object using the current fitted parameters, applied 
        to a completely new dataset that is assumed to be unrelated to the model's
        original data. The new results can then be used for analysis or forecasting.

        Parameters
        ----------
        y : numpy ndarray, pandas Series, pandas DataFrame
            New observations from the modeled time-series process.
        exog : numpy ndarray, pandas Series, pandas DataFrame, default `None`
            New observations of exogenous regressors, if applicable. Must have 
            the same number of observations as `y` and their indexes must be 
            aligned so that y[i] is regressed on exog[i].
        refit : bool, default `False`
            Whether to re-fit the parameters, using the new dataset.
        copy_initialization : bool, default `False`
            Whether or not to copy the initialization from the current results 
            set to the new model. 
        **kwargs
            Keyword arguments may be used to modify model specification arguments 
            when created the new model object.

        Returns
        -------
        None

        Notes
        -----
        The `y` argument to this method should consist of new observations that 
        are not necessarily related to the original model's `y` dataset. For 
        observations that continue that original dataset by follow directly after 
        its last element, see the append and extend methods.

        https://www.statsmodels.org/dev/generated/statsmodels.tsa.statespace.mlemodel.MLEResults.apply.html#statsmodels.tsa.statespace.mlemodel.MLEResults.apply

        """

        fit_kwargs = self._fit_kwargs if refit else None

        self.sarimax_res = self.sarimax_res.apply(
                               endog               = y,
                               exog                = exog,
                               refit               = refit,
                               copy_initialization = copy_initialization,
                               fit_kwargs          = fit_kwargs,
                               **kwargs
                           )


    @_check_fitted
    def extend(
        self,
        y: Union[np.ndarray, pd.Series, pd.DataFrame],
        exog: Optional[Union[np.ndarray, pd.Series, pd.DataFrame]]=None,
        **kwargs
    ) -> None:
        """
        Recreate the results object for new data that extends the original data.

        Creates a new result object applied to a new dataset that is assumed to 
        follow directly from the end of the model's original data. The new 
        results can then be used for analysis or forecasting.

        Parameters
        ----------
        y : numpy ndarray, pandas Series, pandas DataFrame
            New observations from the modeled time-series process.
        exog : numpy ndarray, pandas Series, pandas DataFrame, default `None`
            New observations of exogenous regressors, if applicable. Must have 
            the same number of observations as `y` and their indexes must be 
            aligned so that y[i] is regressed on exog[i].
        **kwargs
            Keyword arguments may be used to modify model specification arguments 
            when created the new model object.

        Returns
        -------
        None

        Notes
        -----
        The `y` argument to this method should consist of new observations that 
        occurred directly after the last element of the model's original `y` 
        array. For any other kind of dataset, see the apply method.

        This method will apply filtering only to the new data provided by the `y` 
        argument, which can be much faster than re-filtering the entire dataset. 
        However, the returned results object will only have results for the new 
        data. To retrieve results for both the new data and the original data, 
        see the append method.

        https://www.statsmodels.org/dev/generated/statsmodels.tsa.statespace.mlemodel.MLEResults.extend.html#statsmodels.tsa.statespace.mlemodel.MLEResults.extend

        """

        self.sarimax_res = self.sarimax_res.extend(
                               endog = y,
                               exog  = exog,
                               **kwargs
                           )


    def set_params(
        self, 
        **params: dict
    ) -> None:
        """
        Set new values to the parameters of the regressor.
        
        Parameters
        ----------
        params : dict
            Parameters values.

        Returns
        -------
        None

        """
        
        params = {k:v for k,v in params.items() if k in self._sarimax_params}
        for key, value in params.items():
            setattr(self, key, value)

        self._consolidate_kwargs()

        # Reset values in case the model has already been fitted.
        self.output_type    = None
        self.sarimax_res    = None
        self.fitted         = False
        self.training_index = None


    @_check_fitted
    def params(
        self
    ) -> Union[np.ndarray, pd.Series]:
        """
        Get the parameters of the model. The order of variables is the trend
        coefficients, the `k_exog` exogenous coefficients, the `k_ar` AR 
        coefficients, and finally the `k_ma` MA coefficients.

        Returns
        -------
        params : numpy ndarray, pandas Series
            The parameters of the model.
        
        """

        return self.sarimax_res.params


    @_check_fitted
    def summary(
        self,
        alpha: float = 0.05,
        start: int = None
    ) -> object:
        """
        Get a summary of the SARIMAXResults object.
        
        Parameters
        ----------
        alpha : float, default `0.05`
            The confidence intervals for the forecasts are (1 - alpha) %.
        start : int, default `None`
            Integer of the start observation.

        Returns
        -------
        summary : Summary instance
            This holds the summary table and text, which can be printed or 
            converted to various output formats.
        
        """

        return self.sarimax_res.summary(alpha=alpha, start=start)


    @_check_fitted
    def get_info_criteria(
        self,
        criteria: str = 'aic',
        method: str = 'standard'
    ) -> float:
        """
        Get the selected information criteria.

        Check https://www.statsmodels.org/dev/generated/statsmodels.tsa.statespace.sarimax.SARIMAXResults.info_criteria.html
        to know more about statsmodels info_criteria method.

        Parameters
        ----------
        criteria : str, default `'aic'`
            The information criteria to compute. Valid options are {'aic', 'bic',
            'hqic'}.
        method : str, default `'standard'`
            The method for information criteria computation. Default is 'standard'
            method; 'lutkepohl' computes the information criteria as in Lütkepohl
            (2007).

        Returns
        -------
        metric : float
            The value of the selected information criteria.
        
        """

        if criteria not in ['aic', 'bic', 'hqic']:
            raise ValueError(
                (f"Invalid value for `criteria`. Valid options are 'aic', 'bic', "
                 f"and 'hqic'.")
            )
        
        if method not in ['standard', 'lutkepohl']:
            raise ValueError(
                (f"Invalid value for `method`. Valid options are 'standard' and "
                 f"'lutkepohl'.")
            )
        
        metric = self.sarimax_res.info_criteria(criteria=criteria, method=method)
        
        return metric


# Unit test set_out_sample_residuals ForecasterAutoregCustom
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.ForecasterAutoregCustom import ForecasterAutoregCustom
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler


def create_predictors(y): # pragma: no cover
    """
    Create first 5 lags of a time series.
    """

    lags = y[-1:-6:-1]

    return lags  


def test_set_out_sample_residuals_TypeError_when_residuals_is_not_numpy_ndarray():
    """
    Test TypeError is raised when residuals argument is not numpy ndarray.
    """
    forecaster = ForecasterAutoregCustom(
                     regressor      = LinearRegression(),
                     fun_predictors = create_predictors,
                     window_size    = 5
                 )
    residuals=pd.Series([1, 2, 3])

    err_msg = re.escape(
                f"`residuals` argument must be `numpy ndarray`. Got {type(residuals)}."
            )
    with pytest.raises(TypeError, match = err_msg):
        forecaster.set_out_sample_residuals(residuals=residuals)


def test_set_out_sample_residuals_warning_when_forecaster_has_transformer_and_transform_False():
    """
    Test Warning is raised when forecaster has a transformer_y and transform=False.
    """
    forecaster = ForecasterAutoregCustom(
                     regressor      = LinearRegression(),
                     fun_predictors = create_predictors,
                     window_size    = 5,
                     transformer_y  = StandardScaler()
                 )
    residuals = np.arange(10)

    err_msg = re.escape(
                (f"Argument `transform` is set to `False` but forecaster was trained "
                 f"using a transformer {forecaster.transformer_y}. Ensure that the new residuals "
                 f"are already transformed or set `transform=True`.")
            )
    with pytest.warns(UserWarning, match = err_msg):
        forecaster.set_out_sample_residuals(residuals=residuals, transform=False)


def test_set_out_sample_residuals_warning_when_forecaster_has_transformer_and_transform_True():
    """
    Test Warning is raised when forecaster has a transformer_y and transform=True.
    """
    forecaster = ForecasterAutoregCustom(
                     regressor      = LinearRegression(),
                     fun_predictors = create_predictors,
                     window_size    = 5,
                     transformer_y  = StandardScaler()
                 )
    forecaster.fit(y=pd.Series(np.arange(10)))
    residuals = np.arange(10)

    err_msg = re.escape(
                (f"Residuals will be transformed using the same transformer used "
                 f"when training the forecaster ({forecaster.transformer_y}). Ensure that the "
                 f"new residuals are on the same scale as the original time series.")
            )
    with pytest.warns(UserWarning, match = err_msg):
        forecaster.set_out_sample_residuals(residuals=residuals, transform=True)


def test_set_out_sample_residuals_when_residuals_length_is_greater_than_1000():
    """
    Test len residuals stored when its length is greater than 1000.
    """
    forecaster = ForecasterAutoregCustom(
                    regressor      = LinearRegression(),
                    fun_predictors = create_predictors,
                    window_size    = 5
                )
    forecaster.set_out_sample_residuals(residuals=np.arange(2000))

    assert len(forecaster.out_sample_residuals) == 1000


def test_same_out_sample_residuals_stored_when_residuals_length_is_greater_than_1000():
    """
    Test same out sample residuals are stored when its length is greater than 1000
    executing function two times.
    """
    forecaster = ForecasterAutoregCustom(
                    regressor      = LinearRegression(),
                    fun_predictors = create_predictors,
                    window_size    = 5
                )
    forecaster.set_out_sample_residuals(residuals=np.arange(2000))
    out_sample_residuals_1 = forecaster.out_sample_residuals
    forecaster = ForecasterAutoregCustom(
                    regressor      = LinearRegression(),
                    fun_predictors = create_predictors,
                    window_size    = 5
                )
    forecaster.set_out_sample_residuals(residuals=np.arange(2000))
    out_sample_residuals_2 = forecaster.out_sample_residuals

    assert (out_sample_residuals_1 == out_sample_residuals_2).all()


def test_set_out_sample_residuals_when_residuals_length_is_less_than_1000_and_no_append():
    """
    Test residuals stored when new residuals length is less than 1000 and append is False.
    """
    forecaster = ForecasterAutoregCustom(
                    regressor      = LinearRegression(),
                    fun_predictors = create_predictors,
                    window_size    = 5
                )
    forecaster.set_out_sample_residuals(residuals=np.arange(20))
    forecaster.set_out_sample_residuals(residuals=np.arange(10), append=False)
    expected = np.arange(10)
    results = forecaster.out_sample_residuals

    assert (results == expected).all()


def test_set_out_sample_residuals_when_residuals_length_is_less_than_1000_and_append():
    """
    Test residuals stored when new residuals length is less than 1000 and append is True.
    """
    forecaster = ForecasterAutoregCustom(
                    regressor      = LinearRegression(),
                    fun_predictors = create_predictors,
                    window_size    = 5
                )
    forecaster.set_out_sample_residuals(residuals=np.arange(10))
    forecaster.set_out_sample_residuals(residuals=np.arange(10), append=True)
    expected = np.hstack([np.arange(10), np.arange(10)])
    results = forecaster.out_sample_residuals

    assert (results == expected).all()


def test_set_out_sample_residuals_when_residuals_length_is_more_than_1000_and_append():
    """
    Test residuals stored when new residuals length is more than 1000 and append is True.
    """
    forecaster = ForecasterAutoregCustom(
                    regressor      = LinearRegression(),
                    fun_predictors = create_predictors,
                    window_size    = 5
                )
    forecaster.set_out_sample_residuals(residuals=np.arange(10))
    forecaster.set_out_sample_residuals(residuals=np.arange(1000), append=True)
    expected = np.hstack([np.arange(10), np.arange(1200)])[:1000]
    results = forecaster.out_sample_residuals

    assert (results == expected).all()


def test_set_out_sample_residuals_when_transform_is_True():
    """
    Test residuals stored when using a StandardScaler() as transformer.
    """

    forecaster = ForecasterAutoregCustom(
                    regressor      = LinearRegression(),
                    fun_predictors = create_predictors,
                    window_size    = 5,
                    transformer_y=StandardScaler()
                    
                )
    y = pd.Series(
            np.array([
                12.5, 10.3,  9.9, 10.4,  9.9,  8.5, 10.6, 11.4, 10. ,  9.5, 10.1,
                11.5, 11.4, 11.3, 10.5,  9.6, 10.4, 11.7,  8.7, 10.6])
        )
    forecaster.fit(y=y)
    new_residuals = np.random.normal(size=100)
    new_residuals_transformed = forecaster.transformer_y.transform(new_residuals.reshape(-1, 1))
    new_residuals_transformed = new_residuals_transformed.flatten()
    forecaster.set_out_sample_residuals(residuals=new_residuals, transform=True)

    np.testing.assert_array_equal(new_residuals_transformed, forecaster.out_sample_residuals)


# Unit test _create_backtesting_folds
# ==============================================================================
import pytest
import numpy as np
import pandas as pd
from skforecast.model_selection.model_selection import _create_backtesting_folds


@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[range(0, 70), range(67, 70), range(70, 80), range(70, 80), False],
                                  [range(0, 70), range(77, 80), range(80, 90), range(80, 90), False],
                                  [range(0, 70), range(87, 90), range(90, 100), range(90, 100), False]]),
                          (False, [[[0, 70], [67, 70], [70, 80], [70, 80], False],
                                   [[0, 70], [77, 80], [80, 90], [80, 90], False],
                                   [[0, 70], [87, 90], [90, 100], [90, 100], False]])], 
                         ids = lambda argument : f'{argument}' )
def test_create_backtesting_folds_no_refit_no_gap_no_remainder(capfd, return_all_indexes, expected):
    """
    Test _create_backtesting_folds output when refit is False, gap=0 and not 
    remainder.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    window_size = 3
    initial_train_size = 70
    test_size = 10
    refit = False

    folds = _create_backtesting_folds(
                data                  = y,
                window_size           = window_size,
                initial_train_size    = initial_train_size,
                test_size             = test_size,
                externally_fitted     = False,
                refit                 = refit,
                fixed_train_size      = False,
                gap                   = 0,
                allow_incomplete_fold = True,
                return_all_indexes    = return_all_indexes,
                verbose               = True
            )
    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of backtesting process\n"
        "----------------------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 3\n"
        "    Number of steps per fold: 10\n"
        "    Number of steps to exclude from the end of each train set before test (gap): 0\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-12 00:00:00 -- 2022-03-21 00:00:00  (n=10)\n"
        "Fold: 1\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-22 00:00:00 -- 2022-03-31 00:00:00  (n=10)\n"
        "Fold: 2\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-04-01 00:00:00 -- 2022-04-10 00:00:00  (n=10)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[range(0, 65), range(61, 65), range(65, 75), range(65, 75), False],
                                  [range(0, 65), range(71, 75), range(75, 85), range(75, 85), False],
                                  [range(0, 65), range(81, 85), range(85, 95), range(85, 95), False]]),
                          (False, [[[0, 65], [61, 65], [65, 75], [65, 75], False],
                                   [[0, 65], [71, 75], [75, 85], [75, 85], False],
                                   [[0, 65], [81, 85], [85, 95], [85, 95], False]])], 
                         ids = lambda argument : f'{argument}' )
def test_create_backtesting_folds_no_refit_no_gap_allow_incomplete_fold_False(capfd, return_all_indexes, expected):
    """
    Test _create_backtesting_folds output when refit is 0 (False), gap=0, 
    remainder and allow_incomplete_fold=False.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    window_size = 4
    initial_train_size = 65
    test_size = 10
    refit = 0
    allow_incomplete_fold = False

    folds = _create_backtesting_folds(
                data                  = y,
                window_size           = window_size,
                initial_train_size    = initial_train_size,
                test_size             = test_size,
                externally_fitted     = False,
                refit                 = refit,
                fixed_train_size      = False,
                gap                   = 0,
                allow_incomplete_fold = allow_incomplete_fold,
                return_all_indexes    = return_all_indexes,
                verbose               = True
            )
    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of backtesting process\n"
        "----------------------------------\n"
        "Number of observations used for initial training: 65\n"
        "Number of observations used for backtesting: 35\n"
        "    Number of folds: 3\n"
        "    Number of steps per fold: 10\n"
        "    Number of steps to exclude from the end of each train set before test (gap): 0\n"
        "    Last fold has been excluded because it was incomplete.\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-06 00:00:00  (n=65)\n"
        "    Validation: 2022-03-07 00:00:00 -- 2022-03-16 00:00:00  (n=10)\n"
        "Fold: 1\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-06 00:00:00  (n=65)\n"
        "    Validation: 2022-03-17 00:00:00 -- 2022-03-26 00:00:00  (n=10)\n"
        "Fold: 2\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-06 00:00:00  (n=65)\n"
        "    Validation: 2022-03-27 00:00:00 -- 2022-04-05 00:00:00  (n=10)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[range(0, 70), range(65, 70), range(70, 82), range(75, 82), False],
                                  [range(0, 70), range(72, 77), range(77, 89), range(82, 89), False],
                                  [range(0, 70), range(79, 84), range(84, 96), range(89, 96), False],
                                  [range(0, 70), range(86, 91), range(91, 100), range(96, 100), False]]),
                          (False, [[[0, 70], [65, 70], [70, 82], [75, 82], False],
                                   [[0, 70], [72, 77], [77, 89], [82, 89], False],
                                   [[0, 70], [79, 84], [84, 96], [89, 96], False],
                                   [[0, 70], [86, 91], [91, 100], [96, 100], False]])], 
                         ids = lambda argument : f'{argument}' )
def test_create_backtesting_folds_no_refit_gap_allow_incomplete_fold_True(capfd, return_all_indexes, expected):
    """
    Test _create_backtesting_folds output when refit is False, gap=5, 
    remainder, allow_incomplete_fold=True.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    window_size = 5
    initial_train_size = 70
    gap = 5
    test_size = 7
    refit = False
    allow_incomplete_fold = True

    folds = _create_backtesting_folds(
                data                  = y,
                window_size           = window_size,
                initial_train_size    = initial_train_size,
                test_size             = test_size,
                externally_fitted     = False,
                refit                 = refit,
                fixed_train_size      = False,
                gap                   = gap,
                allow_incomplete_fold = allow_incomplete_fold,
                return_all_indexes    = return_all_indexes,
                verbose               = True
            )
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of backtesting process\n"
        "----------------------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 4\n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude from the end of each train set before test (gap): 5\n"
        "    Last fold only includes 4 observations.\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-17 00:00:00 -- 2022-03-23 00:00:00  (n=7)\n"
        "Fold: 1\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-24 00:00:00 -- 2022-03-30 00:00:00  (n=7)\n"
        "Fold: 2\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-31 00:00:00 -- 2022-04-06 00:00:00  (n=7)\n"
        "Fold: 3\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-04-07 00:00:00 -- 2022-04-10 00:00:00  (n=4)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[range(0, 15), range(0, 15), range(15, 30), range(20, 30), False],
                                  [range(0, 15), range(10, 25), range(25, 40), range(30, 40), False],
                                  [range(0, 15), range(20, 35), range(35, 50), range(40, 50), False],
                                  [range(0, 15), range(30, 45), range(45, 60), range(50, 60), False],
                                  [range(0, 15), range(40, 55), range(55, 70), range(60, 70), False],
                                  [range(0, 15), range(50, 65), range(65, 80), range(70, 80), False],
                                  [range(0, 15), range(60, 75), range(75, 90), range(80, 90), False],
                                  [range(0, 15), range(70, 85), range(85, 100), range(90, 100), False]]),
                          (False, [[[0, 15], [0, 15], [15, 30], [20, 30], False],
                                   [[0, 15], [10, 25], [25, 40], [30, 40], False],
                                   [[0, 15], [20, 35], [35, 50], [40, 50], False],
                                   [[0, 15], [30, 45], [45, 60], [50, 60], False],
                                   [[0, 15], [40, 55], [55, 70], [60, 70], False],
                                   [[0, 15], [50, 65], [65, 80], [70, 80], False],
                                   [[0, 15], [60, 75], [75, 90], [80, 90], False],
                                   [[0, 15], [70, 85], [85, 100], [90, 100], False]])], 
                         ids = lambda argument : f'{argument}' )
def test_create_backtesting_folds_no_refit_no_initial_train_size_gap(capfd, return_all_indexes, expected):
    """
    Test _create_backtesting_folds output when refit is False, gap=5, 
    already trained, no remainder, allow_incomplete_fold=True.
    """
    y = pd.Series(np.arange(100))
    initial_train_size = 15 # window_size
    window_size = 15
    gap = 5
    test_size = 10
    refit = False
    externally_fitted = True

    folds = _create_backtesting_folds(
                data                  = y,
                window_size           = window_size,
                initial_train_size    = initial_train_size,
                test_size             = test_size,
                externally_fitted     = externally_fitted,
                refit                 = refit,
                fixed_train_size      = False,
                gap                   = gap,
                allow_incomplete_fold = True,
                return_all_indexes    = return_all_indexes,
                verbose               = True
            )
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of backtesting process\n"
        "----------------------------------\n"
        "An already trained forecaster is to be used. Window size: 15\n"
        "Number of observations used for backtesting: 85\n"
        "    Number of folds: 8\n"
        "    Number of steps per fold: 10\n"
        "    Number of steps to exclude from the end of each train set before test (gap): 5\n\n"
        "Fold: 0\n"
        "    Validation: 20 -- 29  (n=10)\n"
        "Fold: 1\n"
        "    Validation: 30 -- 39  (n=10)\n"
        "Fold: 2\n"
        "    Validation: 40 -- 49  (n=10)\n"
        "Fold: 3\n"
        "    Validation: 50 -- 59  (n=10)\n"
        "Fold: 4\n"
        "    Validation: 60 -- 69  (n=10)\n"
        "Fold: 5\n"
        "    Validation: 70 -- 79  (n=10)\n"
        "Fold: 6\n"
        "    Validation: 80 -- 89  (n=10)\n"
        "Fold: 7\n"
        "    Validation: 90 -- 99  (n=10)\n\n"
    )

    assert out == expected_out
    assert folds == expected


@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[range(0, 70), range(68, 70), range(70, 80), range(70, 80), False],
                                  [range(0, 80), range(78, 80), range(80, 90), range(80, 90), True],
                                  [range(0, 90), range(88, 90), range(90, 100), range(90, 100), True]]),
                          (False, [[[0, 70], [68, 70], [70, 80], [70, 80], False],
                                   [[0, 80], [78, 80], [80, 90], [80, 90], True],
                                   [[0, 90], [88, 90], [90, 100], [90, 100], True]])], 
                         ids = lambda argument : f'{argument}' )
def test_create_backtesting_folds_refit_no_fixed_no_gap_no_remainder(capfd, return_all_indexes, expected):
    """
    Test _create_backtesting_folds output when refit is True, fixed_train_size is 
    False, gap=0 and not remainder.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    window_size = 2
    initial_train_size = 70
    test_size = 10
    refit = True
    fixed_train_size = False

    folds = _create_backtesting_folds(
                data                  = y,
                window_size           = window_size,
                initial_train_size    = initial_train_size,
                test_size             = test_size,
                externally_fitted     = False,
                refit                 = refit,
                fixed_train_size      = fixed_train_size,
                gap                   = 0,
                allow_incomplete_fold = True,
                return_all_indexes    = return_all_indexes,
                verbose               = True
            )
    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of backtesting process\n"
        "----------------------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 3\n"
        "    Number of steps per fold: 10\n"
        "    Number of steps to exclude from the end of each train set before test (gap): 0\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-12 00:00:00 -- 2022-03-21 00:00:00  (n=10)\n"
        "Fold: 1\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-21 00:00:00  (n=80)\n"
        "    Validation: 2022-03-22 00:00:00 -- 2022-03-31 00:00:00  (n=10)\n"
        "Fold: 2\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-31 00:00:00  (n=90)\n"
        "    Validation: 2022-04-01 00:00:00 -- 2022-04-10 00:00:00  (n=10)\n\n"
    )

    assert out == expected_out
    assert folds == expected


@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[range(0, 70), range(64, 70), range(70, 80), range(70, 80), False],
                                  [range(10, 80), range(74, 80), range(80, 90), range(80, 90), True],
                                  [range(20, 90), range(84, 90), range(90, 100), range(90, 100), True]]),
                          (False, [[[0, 70], [64, 70], [70, 80], [70, 80], False],
                                   [[10, 80], [74, 80], [80, 90], [80, 90], True],
                                   [[20, 90], [84, 90], [90, 100], [90, 100], True]])], 
                         ids = lambda argument : f'{argument}' )
def test_create_backtesting_folds_refit_fixed_train_size_no_gap_no_remainder(capfd, return_all_indexes, expected):
    """
    Test _create_backtesting_folds output when refit is 1 (True), fixed_train_size 
    is True, gap=0 and not remainder.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    window_size = 6
    initial_train_size = 70
    test_size = 10
    refit = 1
    fixed_train_size = True

    folds = _create_backtesting_folds(
                data                  = y,
                window_size           = window_size,
                initial_train_size    = initial_train_size,
                test_size             = test_size,
                externally_fitted     = False,
                refit                 = refit,
                fixed_train_size      = fixed_train_size,
                gap                   = 0,
                allow_incomplete_fold = True,
                return_all_indexes    = return_all_indexes,
                verbose               = True
            )
    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of backtesting process\n"
        "----------------------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 3\n"
        "    Number of steps per fold: 10\n"
        "    Number of steps to exclude from the end of each train set before test (gap): 0\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-12 00:00:00 -- 2022-03-21 00:00:00  (n=10)\n"
        "Fold: 1\n"
        "    Training:   2022-01-11 00:00:00 -- 2022-03-21 00:00:00  (n=70)\n"
        "    Validation: 2022-03-22 00:00:00 -- 2022-03-31 00:00:00  (n=10)\n"
        "Fold: 2\n"
        "    Training:   2022-01-21 00:00:00 -- 2022-03-31 00:00:00  (n=70)\n"
        "    Validation: 2022-04-01 00:00:00 -- 2022-04-10 00:00:00  (n=10)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[range(0, 70), range(67, 70), range(70, 82), range(75, 82), False],
                                  [range(0, 77), range(74, 77), range(77, 89), range(82, 89), True],
                                  [range(0, 84), range(81, 84), range(84, 96), range(89, 96), True],
                                  [range(0, 91), range(88, 91), range(91, 100), range(96, 100), True]]),
                          (False, [[[0, 70], [67, 70], [70, 82], [75, 82], False],
                                   [[0, 77], [74, 77], [77, 89], [82, 89], True],
                                   [[0, 84], [81, 84], [84, 96], [89, 96], True],
                                   [[0, 91], [88, 91], [91, 100], [96, 100], True]])], 
                         ids = lambda argument : f'{argument}' )
def test_create_backtesting_folds_refit_no_fixed_gap_allow_incomplete_fold_True(capfd, return_all_indexes, expected):
    """
    Test _create_backtesting_folds output when refit is True, fixed_train_size is 
    False, gap=5, remainder, allow_incomplete_fold=True.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    window_size = 3
    initial_train_size = 70
    gap = 5
    test_size = 7
    refit = True
    fixed_train_size = False
    allow_incomplete_fold = True

    folds = _create_backtesting_folds(
                data                  = y,
                window_size           = window_size,
                initial_train_size    = initial_train_size,
                test_size             = test_size,
                externally_fitted     = False,
                refit                 = refit,
                fixed_train_size      = fixed_train_size,
                gap                   = gap,
                allow_incomplete_fold = allow_incomplete_fold,
                return_all_indexes    = return_all_indexes,
                verbose               = True
            )
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of backtesting process\n"
        "----------------------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 4\n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude from the end of each train set before test (gap): 5\n"
        "    Last fold only includes 4 observations.\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-17 00:00:00 -- 2022-03-23 00:00:00  (n=7)\n"
        "Fold: 1\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-18 00:00:00  (n=77)\n"
        "    Validation: 2022-03-24 00:00:00 -- 2022-03-30 00:00:00  (n=7)\n"
        "Fold: 2\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-25 00:00:00  (n=84)\n"
        "    Validation: 2022-03-31 00:00:00 -- 2022-04-06 00:00:00  (n=7)\n"
        "Fold: 3\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-04-01 00:00:00  (n=91)\n"
        "    Validation: 2022-04-07 00:00:00 -- 2022-04-10 00:00:00  (n=4)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[range(0, 70), range(65, 70), range(70, 82), range(75, 82), False],
                                  [range(7, 77), range(72, 77), range(77, 89), range(82, 89), True],
                                  [range(14, 84), range(79, 84), range(84, 96), range(89, 96), True],
                                  [range(21, 91), range(86, 91), range(91, 100), range(96, 100), True]]),
                          (False, [[[0, 70], [65, 70], [70, 82], [75, 82], False],
                                   [[7, 77], [72, 77], [77, 89], [82, 89], True],
                                   [[14, 84], [79, 84], [84, 96], [89, 96], True],
                                   [[21, 91], [86, 91], [91, 100], [96, 100], True]])], 
                         ids = lambda argument : f'{argument}' )
def test_create_backtesting_folds_refit_fixed_train_size_gap_allow_incomplete_fold_True(capfd, return_all_indexes, expected):
    """
    Test _create_backtesting_folds output when refit is True, fixed_train_size is 
    True, gap=5, remainder, allow_incomplete_fold=True.
    """
    y = pd.Series(np.arange(100))
    window_size = 5
    initial_train_size = 70
    gap = 5
    test_size = 7
    refit = True
    fixed_train_size = True
    allow_incomplete_fold = True

    folds = _create_backtesting_folds(
                data                  = y,
                window_size           = window_size,
                initial_train_size    = initial_train_size,
                test_size             = test_size,
                externally_fitted     = False,
                refit                 = refit,
                fixed_train_size      = fixed_train_size,
                gap                   = gap,
                allow_incomplete_fold = allow_incomplete_fold,
                return_all_indexes    = return_all_indexes,
                verbose               = True
            )
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of backtesting process\n"
        "----------------------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 4\n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude from the end of each train set before test (gap): 5\n"
        "    Last fold only includes 4 observations.\n\n"
        "Fold: 0\n"
        "    Training:   0 -- 69  (n=70)\n"
        "    Validation: 75 -- 81  (n=7)\n"
        "Fold: 1\n"
        "    Training:   7 -- 76  (n=70)\n"
        "    Validation: 82 -- 88  (n=7)\n"
        "Fold: 2\n"
        "    Training:   14 -- 83  (n=70)\n"
        "    Validation: 89 -- 95  (n=7)\n"
        "Fold: 3\n"
        "    Training:   21 -- 90  (n=70)\n"
        "    Validation: 96 -- 99  (n=4)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[range(0, 70), range(67, 70), range(70, 82), range(75, 82), False],
                                  [range(0, 77), range(74, 77), range(77, 89), range(82, 89), True],
                                  [range(0, 84), range(81, 84), range(84, 96), range(89, 96), True]]),
                          (False, [[[0, 70], [67, 70], [70, 82], [75, 82], False],
                                   [[0, 77], [74, 77], [77, 89], [82, 89], True],
                                   [[0, 84], [81, 84], [84, 96], [89, 96], True]])], 
                         ids = lambda argument : f'{argument}' )
def test_create_backtesting_folds_refit_no_fixed_gap_allow_incomplete_fold_False(capfd, return_all_indexes, expected):
    """
    Test _create_backtesting_folds output when refit is True, fixed_train_size is 
    False, gap=5, remainder, allow_incomplete_fold=False.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    window_size = 3
    initial_train_size = 70
    gap = 5
    test_size = 7
    refit = True
    fixed_train_size = False
    allow_incomplete_fold = False

    folds = _create_backtesting_folds(
                data                  = y,
                window_size           = window_size,
                initial_train_size    = initial_train_size,
                test_size             = test_size,
                externally_fitted     = False,
                refit                 = refit,
                fixed_train_size      = fixed_train_size,
                gap                   = gap,
                allow_incomplete_fold = allow_incomplete_fold,
                return_all_indexes    = return_all_indexes,
                verbose               = True
            )
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of backtesting process\n"
        "----------------------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 3\n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude from the end of each train set before test (gap): 5\n"
        "    Last fold has been excluded because it was incomplete.\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-17 00:00:00 -- 2022-03-23 00:00:00  (n=7)\n"
        "Fold: 1\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-18 00:00:00  (n=77)\n"
        "    Validation: 2022-03-24 00:00:00 -- 2022-03-30 00:00:00  (n=7)\n"
        "Fold: 2\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-25 00:00:00  (n=84)\n"
        "    Validation: 2022-03-31 00:00:00 -- 2022-04-06 00:00:00  (n=7)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[range(0, 70), range(66, 70), range(70, 82), range(75, 82), False],
                                  [range(7, 77), range(73, 77), range(77, 89), range(82, 89), True],
                                  [range(14, 84), range(80, 84), range(84, 96), range(89, 96), True]]),
                          (False, [[[0, 70], [66, 70], [70, 82], [75, 82], False],
                                   [[7, 77], [73, 77], [77, 89], [82, 89], True],
                                   [[14, 84], [80, 84], [84, 96], [89, 96], True]])], 
                         ids = lambda argument : f'{argument}' )
def test_create_backtesting_folds_refit_fixed_train_size_gap_allow_incomplete_fold_False(capfd, return_all_indexes, expected):
    """
    Test _create_backtesting_folds output when refit is True, fixed_train_size is 
    True, gap=5, remainder, allow_incomplete_fold=False.
    """
    y = pd.Series(np.arange(100))
    window_size = 4
    initial_train_size = 70
    gap = 5
    test_size = 7
    refit = True
    fixed_train_size = True
    allow_incomplete_fold = False

    folds = _create_backtesting_folds(
                data                  = y,
                window_size           = window_size,
                initial_train_size    = initial_train_size,
                test_size             = test_size,
                externally_fitted     = False,
                refit                 = refit,
                fixed_train_size      = fixed_train_size,
                gap                   = gap,
                allow_incomplete_fold = allow_incomplete_fold,
                return_all_indexes    = return_all_indexes,
                verbose               = True
            )
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of backtesting process\n"
        "----------------------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 3\n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude from the end of each train set before test (gap): 5\n"
        "    Last fold has been excluded because it was incomplete.\n\n"
        "Fold: 0\n"
        "    Training:   0 -- 69  (n=70)\n"
        "    Validation: 75 -- 81  (n=7)\n"
        "Fold: 1\n"
        "    Training:   7 -- 76  (n=70)\n"
        "    Validation: 82 -- 88  (n=7)\n"
        "Fold: 2\n"
        "    Training:   14 -- 83  (n=70)\n"
        "    Validation: 89 -- 95  (n=7)\n\n"
    )

    assert out == expected_out
    assert folds == expected


@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[range(0, 60), range(56, 60), range(60, 70), range(60, 70), False],
                                  [range(0, 60), range(66, 70), range(70, 80), range(70, 80), False],
                                  [range(0, 80), range(76, 80), range(80, 90), range(80, 90), True],
                                  [range(0, 80), range(86, 90), range(90, 100), range(90, 100), False]]),
                          (False, [[[0, 60], [56, 60], [60, 70], [60, 70], False],
                                   [[0, 60], [66, 70], [70, 80], [70, 80], False],
                                   [[0, 80], [76, 80], [80, 90], [80, 90], True],
                                   [[0, 80], [86, 90], [90, 100], [90, 100], False]])], 
                         ids = lambda argument : f'{argument}' )
def test_create_backtesting_folds_refit_int_no_fixed_no_gap_no_remainder(capfd, return_all_indexes, expected):
    """
    Test _create_backtesting_folds output when refit is 2, fixed_train_size is 
    False, gap=0 and not remainder.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    window_size = 4
    initial_train_size = 60
    test_size = 10
    refit = 2
    fixed_train_size = False

    folds = _create_backtesting_folds(
                data                  = y,
                window_size           = window_size,
                initial_train_size    = initial_train_size,
                test_size             = test_size,
                externally_fitted     = False,
                refit                 = refit,
                fixed_train_size      = fixed_train_size,
                gap                   = 0,
                allow_incomplete_fold = True,
                return_all_indexes    = return_all_indexes,
                verbose               = True
            )
    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of backtesting process\n"
        "----------------------------------\n"
        "Number of observations used for initial training: 60\n"
        "Number of observations used for backtesting: 40\n"
        "    Number of folds: 4\n"
        "    Number of steps per fold: 10\n"
        "    Number of steps to exclude from the end of each train set before test (gap): 0\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-01 00:00:00  (n=60)\n"
        "    Validation: 2022-03-02 00:00:00 -- 2022-03-11 00:00:00  (n=10)\n"
        "Fold: 1\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-01 00:00:00  (n=60)\n"
        "    Validation: 2022-03-12 00:00:00 -- 2022-03-21 00:00:00  (n=10)\n"
        "Fold: 2\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-21 00:00:00  (n=80)\n"
        "    Validation: 2022-03-22 00:00:00 -- 2022-03-31 00:00:00  (n=10)\n"
        "Fold: 3\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-21 00:00:00  (n=80)\n"
        "    Validation: 2022-04-01 00:00:00 -- 2022-04-10 00:00:00  (n=10)\n\n"
    )

    assert out == expected_out
    assert folds == expected


@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[range(0, 60), range(58, 60), range(60, 70), range(60, 70), False],
                                  [range(0, 60), range(68, 70), range(70, 80), range(70, 80), False],
                                  [range(0, 60), range(78, 80), range(80, 90), range(80, 90), False],
                                  [range(30, 90), range(88, 90), range(90, 100), range(90, 100), True]]),
                          (False, [[[0, 60], [58, 60], [60, 70], [60, 70], False],
                                   [[0, 60], [68, 70], [70, 80], [70, 80], False],
                                   [[0, 60], [78, 80], [80, 90], [80, 90], False],
                                   [[30, 90], [88, 90], [90, 100], [90, 100], True]])], 
                         ids = lambda argument : f'{argument}' )
def test_create_backtesting_folds_refit_int_fixed_train_size_no_gap_no_remainder(capfd, return_all_indexes, expected):
    """
    Test _create_backtesting_folds output when refit is 3, fixed_train_size is 
    True, gap=0 and not remainder.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    window_size = 2
    initial_train_size = 60
    test_size = 10
    refit = 3
    fixed_train_size = True

    folds = _create_backtesting_folds(
                data                  = y,
                window_size           = window_size,
                initial_train_size    = initial_train_size,
                test_size             = test_size,
                externally_fitted     = False,
                refit                 = refit,
                fixed_train_size      = fixed_train_size,
                gap                   = 0,
                allow_incomplete_fold = True,
                return_all_indexes    = return_all_indexes,
                verbose               = True
            )
    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of backtesting process\n"
        "----------------------------------\n"
        "Number of observations used for initial training: 60\n"
        "Number of observations used for backtesting: 40\n"
        "    Number of folds: 4\n"
        "    Number of steps per fold: 10\n"
        "    Number of steps to exclude from the end of each train set before test (gap): 0\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-01 00:00:00  (n=60)\n"
        "    Validation: 2022-03-02 00:00:00 -- 2022-03-11 00:00:00  (n=10)\n"
        "Fold: 1\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-01 00:00:00  (n=60)\n"
        "    Validation: 2022-03-12 00:00:00 -- 2022-03-21 00:00:00  (n=10)\n"
        "Fold: 2\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-01 00:00:00  (n=60)\n"
        "    Validation: 2022-03-22 00:00:00 -- 2022-03-31 00:00:00  (n=10)\n"
        "Fold: 3\n"
        "    Training:   2022-01-31 00:00:00 -- 2022-03-31 00:00:00  (n=60)\n"
        "    Validation: 2022-04-01 00:00:00 -- 2022-04-10 00:00:00  (n=10)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[range(0, 70), range(60, 70), range(70, 82), range(75, 82), False],
                                  [range(0, 70), range(67, 77), range(77, 89), range(82, 89), False],
                                  [range(0, 84), range(74, 84), range(84, 96), range(89, 96), True],
                                  [range(0, 84), range(81, 91), range(91, 100), range(96, 100), False]]),
                          (False, [[[0, 70], [60, 70], [70, 82], [75, 82], False],
                                   [[0, 70], [67, 77], [77, 89], [82, 89], False],
                                   [[0, 84], [74, 84], [84, 96], [89, 96], True],
                                   [[0, 84], [81, 91], [91, 100], [96, 100], False]])], 
                         ids = lambda argument : f'{argument}' )
def test_create_backtesting_folds_refit_int_no_fixed_gap_allow_incomplete_fold_True(capfd, return_all_indexes, expected):
    """
    Test _create_backtesting_folds output when refit is 2, fixed_train_size is 
    False, gap=5, remainder, allow_incomplete_fold=True.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    window_size = 10
    initial_train_size = 70
    gap = 5
    test_size = 7
    refit = 2
    fixed_train_size = False
    allow_incomplete_fold = True

    folds = _create_backtesting_folds(
                data                  = y,
                window_size           = window_size,
                initial_train_size    = initial_train_size,
                test_size             = test_size,
                externally_fitted     = False,
                refit                 = refit,
                fixed_train_size      = fixed_train_size,
                gap                   = gap,
                allow_incomplete_fold = allow_incomplete_fold,
                return_all_indexes    = return_all_indexes,
                verbose               = True
            )
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of backtesting process\n"
        "----------------------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 4\n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude from the end of each train set before test (gap): 5\n"
        "    Last fold only includes 4 observations.\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-17 00:00:00 -- 2022-03-23 00:00:00  (n=7)\n"
        "Fold: 1\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-24 00:00:00 -- 2022-03-30 00:00:00  (n=7)\n"
        "Fold: 2\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-25 00:00:00  (n=84)\n"
        "    Validation: 2022-03-31 00:00:00 -- 2022-04-06 00:00:00  (n=7)\n"
        "Fold: 3\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-25 00:00:00  (n=84)\n"
        "    Validation: 2022-04-07 00:00:00 -- 2022-04-10 00:00:00  (n=4)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[range(0, 70), range(55, 70), range(70, 82), range(75, 82), False],
                                  [range(0, 70), range(62, 77), range(77, 89), range(82, 89), False],
                                  [range(0, 70), range(69, 84), range(84, 96), range(89, 96), False],
                                  [range(21, 91), range(76, 91), range(91, 100), range(96, 100), True]]),
                          (False, [[[0, 70], [55, 70], [70, 82], [75, 82], False],
                                   [[0, 70], [62, 77], [77, 89], [82, 89], False],
                                   [[0, 70], [69, 84], [84, 96], [89, 96], False],
                                   [[21, 91], [76, 91], [91, 100], [96, 100], True]])], 
                         ids = lambda argument : f'{argument}' )
def test_create_backtesting_folds_refit_int_fixed_train_size_gap_allow_incomplete_fold_True(capfd, return_all_indexes, expected):
    """
    Test _create_backtesting_folds output when refit is 3, fixed_train_size is 
    True, gap=5, remainder, allow_incomplete_fold=True.
    """
    y = pd.Series(np.arange(100))
    window_size = 15
    initial_train_size = 70
    gap = 5
    test_size = 7
    refit = 3
    fixed_train_size = True
    allow_incomplete_fold = True

    folds = _create_backtesting_folds(
                data                  = y,
                window_size           = window_size,
                initial_train_size    = initial_train_size,
                test_size             = test_size,
                externally_fitted     = False,
                refit                 = refit,
                fixed_train_size      = fixed_train_size,
                gap                   = gap,
                allow_incomplete_fold = allow_incomplete_fold,
                return_all_indexes    = return_all_indexes,
                verbose               = True
            )
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of backtesting process\n"
        "----------------------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 4\n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude from the end of each train set before test (gap): 5\n"
        "    Last fold only includes 4 observations.\n\n"
        "Fold: 0\n"
        "    Training:   0 -- 69  (n=70)\n"
        "    Validation: 75 -- 81  (n=7)\n"
        "Fold: 1\n"
        "    Training:   0 -- 69  (n=70)\n"
        "    Validation: 82 -- 88  (n=7)\n"
        "Fold: 2\n"
        "    Training:   0 -- 69  (n=70)\n"
        "    Validation: 89 -- 95  (n=7)\n"
        "Fold: 3\n"
        "    Training:   21 -- 90  (n=70)\n"
        "    Validation: 96 -- 99  (n=4)\n\n"
    )

    assert out == expected_out
    assert folds == expected
    

@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[range(0, 70), range(50, 70), range(70, 82), range(75, 82), False],
                                  [range(0, 70), range(57, 77), range(77, 89), range(82, 89), False],
                                  [range(0, 70), range(64, 84), range(84, 96), range(89, 96), False]]),
                          (False, [[[0, 70], [50, 70], [70, 82], [75, 82], False],
                                   [[0, 70], [57, 77], [77, 89], [82, 89], False],
                                   [[0, 70], [64, 84], [84, 96], [89, 96], False]])], 
                         ids = lambda argument : f'{argument}' )
def test_create_backtesting_folds_refit_int_no_fixed_gap_allow_incomplete_fold_False(capfd, return_all_indexes, expected):
    """
    Test _create_backtesting_folds output when refit is 3, fixed_train_size is 
    False, gap=5, remainder, allow_incomplete_fold=False.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    window_size = 20
    initial_train_size = 70
    gap = 5
    test_size = 7
    refit = 3
    fixed_train_size = False
    allow_incomplete_fold = False

    folds = _create_backtesting_folds(
                data                  = y,
                window_size           = window_size,
                initial_train_size    = initial_train_size,
                test_size             = test_size,
                externally_fitted     = False,
                refit                 = refit,
                fixed_train_size      = fixed_train_size,
                gap                   = gap,
                allow_incomplete_fold = allow_incomplete_fold,
                return_all_indexes    = return_all_indexes,
                verbose               = True
            )
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of backtesting process\n"
        "----------------------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 3\n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude from the end of each train set before test (gap): 5\n"
        "    Last fold has been excluded because it was incomplete.\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-17 00:00:00 -- 2022-03-23 00:00:00  (n=7)\n"
        "Fold: 1\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-24 00:00:00 -- 2022-03-30 00:00:00  (n=7)\n"
        "Fold: 2\n"
        "    Training:   2022-01-01 00:00:00 -- 2022-03-11 00:00:00  (n=70)\n"
        "    Validation: 2022-03-31 00:00:00 -- 2022-04-06 00:00:00  (n=7)\n\n"
    )

    assert out == expected_out
    assert folds == expected


@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[range(0, 70), range(60, 70), range(70, 82), range(75, 82), False],
                                  [range(0, 70), range(67, 77), range(77, 89), range(82, 89), False],
                                  [range(14, 84), range(74, 84), range(84, 96), range(89, 96), True]]),
                          (False, [[[0, 70], [60, 70], [70, 82], [75, 82], False],
                                   [[0, 70], [67, 77], [77, 89], [82, 89], False],
                                   [[14, 84], [74, 84], [84, 96], [89, 96], True]])], 
                         ids = lambda argument : f'{argument}' )
def test_create_backtesting_folds_refit_int_fixed_train_size_gap_allow_incomplete_fold_False(capfd, return_all_indexes, expected):
    """
    Test _create_backtesting_folds output when refit is True, fixed_train_size is 
    True, gap=5, remainder, allow_incomplete_fold=False.
    """
    y = pd.Series(np.arange(100))
    window_size = 10
    initial_train_size = 70
    gap = 5
    test_size = 7
    refit = 2
    fixed_train_size = True
    allow_incomplete_fold = False

    folds = _create_backtesting_folds(
                data                  = y,
                window_size           = window_size,
                initial_train_size    = initial_train_size,
                test_size             = test_size,
                externally_fitted     = False,
                refit                 = refit,
                fixed_train_size      = fixed_train_size,
                gap                   = gap,
                allow_incomplete_fold = allow_incomplete_fold,
                return_all_indexes    = return_all_indexes,
                verbose               = True
            )
                    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of backtesting process\n"
        "----------------------------------\n"
        "Number of observations used for initial training: 70\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 3\n"
        "    Number of steps per fold: 7\n"
        "    Number of steps to exclude from the end of each train set before test (gap): 5\n"
        "    Last fold has been excluded because it was incomplete.\n\n"
        "Fold: 0\n"
        "    Training:   0 -- 69  (n=70)\n"
        "    Validation: 75 -- 81  (n=7)\n"
        "Fold: 1\n"
        "    Training:   0 -- 69  (n=70)\n"
        "    Validation: 82 -- 88  (n=7)\n"
        "Fold: 2\n"
        "    Training:   14 -- 83  (n=70)\n"
        "    Validation: 89 -- 95  (n=7)\n\n"
    )

    assert out == expected_out
    assert folds == expected


@pytest.mark.parametrize("return_all_indexes, expected",
                         [(True, [[range(0, 70), range(67, 70), range(70, 80), range(70, 80), False],
                                  [range(0, 80), range(77, 80), range(80, 90), range(80, 90), True],
                                  [range(0, 90), range(87, 90), range(90, 100), range(90, 100), True]]),
                          (False, [[[0, 70], [67, 70], [70, 80], [70, 80], False],
                                   [[0, 80], [77, 80], [80, 90], [80, 90], True],
                                   [[0, 90], [87, 90], [90, 100], [90, 100], True]])], 
                         ids = lambda argument : f'{argument}' )
def test_create_backtesting_folds_refit_no_fixed_no_gap_no_remainder_differentiation(capfd, return_all_indexes, expected):
    """
    Test _create_backtesting_folds output when refit is True, fixed_train_size is 
    False, gap=0, not remainder and differentiation=1.
    """
    y = pd.Series(np.arange(100))
    y.index = pd.date_range(start='2022-01-01', periods=100, freq='D')
    differentiation = 1
    window_size = 2 + differentiation
    initial_train_size = 70
    test_size = 10
    refit = True
    fixed_train_size = False

    folds = _create_backtesting_folds(
                data                  = y,
                window_size           = window_size,
                initial_train_size    = initial_train_size,
                test_size             = test_size,
                externally_fitted     = False,
                refit                 = refit,
                fixed_train_size      = fixed_train_size,
                gap                   = 0,
                allow_incomplete_fold = True,
                return_all_indexes    = return_all_indexes,
                differentiation       = differentiation,
                verbose               = True
            )
    
    out, _ = capfd.readouterr()
    expected_out = (
        "Information of backtesting process\n"
        "----------------------------------\n"
        "Number of observations used for initial training: 69\n"
        "    First 1 observation/s in training sets are used for differentiation\n"
        "Number of observations used for backtesting: 30\n"
        "    Number of folds: 3\n"
        "    Number of steps per fold: 10\n"
        "    Number of steps to exclude from the end of each train set before test (gap): 0\n\n"
        "Fold: 0\n"
        "    Training:   2022-01-02 00:00:00 -- 2022-03-11 00:00:00  (n=69)\n"
        "    Validation: 2022-03-12 00:00:00 -- 2022-03-21 00:00:00  (n=10)\n"
        "Fold: 1\n"
        "    Training:   2022-01-02 00:00:00 -- 2022-03-21 00:00:00  (n=79)\n"
        "    Validation: 2022-03-22 00:00:00 -- 2022-03-31 00:00:00  (n=10)\n"
        "Fold: 2\n"
        "    Training:   2022-01-02 00:00:00 -- 2022-03-31 00:00:00  (n=89)\n"
        "    Validation: 2022-04-01 00:00:00 -- 2022-04-10 00:00:00  (n=10)\n\n"
    )

    assert out == expected_out
    assert folds == expected


# Unit test _initialize_levels_model_selection_multiseries
# ==============================================================================
import re
import pytest
from sklearn.linear_model import Ridge
from skforecast.ForecasterAutoregMultiSeries import ForecasterAutoregMultiSeries
from skforecast.ForecasterAutoregMultiSeriesCustom import ForecasterAutoregMultiSeriesCustom
from skforecast.ForecasterAutoregMultiVariate import ForecasterAutoregMultiVariate
from skforecast.model_selection_multiseries.model_selection_multiseries import _initialize_levels_model_selection_multiseries
from skforecast.exceptions import IgnoredArgumentWarning

# Fixtures
from .fixtures_model_selection_multiseries import series

def create_predictors(y): # pragma: no cover
    """
    Create first 4 lags of a time series, used in ForecasterAutoregCustom.
    """

    lags = y[-1:-5:-1]

    return lags


@pytest.mark.parametrize("forecaster", 
                         [ForecasterAutoregMultiSeries(
                              regressor = Ridge(random_state=123),
                              lags      = 4),
                          ForecasterAutoregMultiSeriesCustom(
                              regressor      = Ridge(random_state=123),
                              fun_predictors = create_predictors,
                              window_size    = 4)], 
                         ids=lambda fn: f'forecaster_name: {type(fn).__name__}')
def test_initialize_levels_model_selection_multiseries_TypeError_when_levels_not_list_str_None(forecaster):
    """
    Test TypeError is raised in _initialize_levels_model_selection_multiseries when 
    `levels` is not a `list`, `str` or `None`.
    """
    levels = 5
    err_msg = re.escape(
        ("`levels` must be a `list` of column names, a `str` of a column "
         "name or `None` when using a forecaster of type "
         "['ForecasterAutoregMultiSeries', 'ForecasterAutoregMultiSeriesCustom', "
         "'ForecasterRnn']. If the forecaster is of type "
         "`ForecasterAutoregMultiVariate`, this argument is ignored.")
    )
    with pytest.raises(TypeError, match = err_msg):
        _initialize_levels_model_selection_multiseries(
            forecaster = forecaster, 
            series     = series,
            levels     = levels
        )


def test_initialize_levels_model_selection_multiseries_IgnoredArgumentWarning_forecaster_multivariate_and_levels():
    """
    Test IgnoredArgumentWarning is raised when levels is not forecaster.level or 
    None in ForecasterAutoregMultiVariate.
    """
    forecaster = ForecasterAutoregMultiVariate(
                     regressor = Ridge(random_state=123),
                     level     = 'l1',
                     lags      = 2,
                     steps     = 3
                 )
    
    levels = 'not_l1_or_None'
    
    warn_msg = re.escape(
        ("`levels` argument have no use when the forecaster is of type "
         "`ForecasterAutoregMultiVariate`. The level of this forecaster is "
         "'l1', to predict another level, change the `level` "
         "argument when initializing the forecaster.")
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        levels = _initialize_levels_model_selection_multiseries(
                     forecaster = forecaster, 
                     series     = series,
                     levels     = levels
                 )
        
    assert levels == ['l1']


@pytest.mark.parametrize("levels, levels_expected",
                         [(None, ['l1', 'l2']), 
                          ('l1', ['l1']),
                          (['l1', 'l2'], ['l1', 'l2'])],
                         ids=lambda lags: f'lags, lags_grid_expected: {lags}')
def test_initialize_levels_model_selection_multiseries_for_all_inputs(levels, levels_expected):
    """
    Test initialize_levels_model_selection_multiseries when levels is None, 
    str or list.
    """
    forecaster = ForecasterAutoregMultiSeries(
                     regressor = Ridge(random_state=123),
                     lags      = 2
                 )
    
    levels = _initialize_levels_model_selection_multiseries(
                 forecaster = forecaster, 
                 series     = series,
                 levels     = levels
             )
    
    assert levels == levels_expected


@pytest.mark.parametrize("levels, levels_expected",
                         [(None, ['l1', 'l2']), 
                          ('l1', ['l1']),
                          (['l1', 'l2'], ['l1', 'l2'])],
                         ids=lambda lags: f'lags, lags_grid_expected: {lags}')
def test_initialize_levels_model_selection_multiseries_for_all_inputs_series_as_dict(levels, levels_expected):
    """
    Test initialize_levels_model_selection_multiseries when levels is None, 
    str or list when series is a dict.
    """
    series_as_dict = series.to_dict()
    forecaster = ForecasterAutoregMultiSeries(
                     regressor = Ridge(random_state=123),
                     lags      = 2
                 )
    
    levels = _initialize_levels_model_selection_multiseries(
                 forecaster = forecaster, 
                 series     = series_as_dict,
                 levels     = levels
             )
    
    assert levels == levels_expected


# Unit test get_feature_importances ForecasterAutoregMultiSeries
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.ForecasterAutoregMultiSeries import ForecasterAutoregMultiSeries
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LinearRegression
from sklearn.neural_network import MLPRegressor
from sklearn.ensemble import RandomForestRegressor

# Fixtures
series = pd.DataFrame({'1': pd.Series(np.arange(10)), 
                       '2': pd.Series(np.arange(10))})


def test_NotFittedError_is_raised_when_forecaster_is_not_fitted():
    """
    Test NotFittedError is raised when calling get_feature_importances() and 
    forecaster is not fitted.
    """
    forecaster = ForecasterAutoregMultiSeries(
                     regressor = LinearRegression(),
                     lags      = 3,
                 )

    err_msg = re.escape(
        ("This forecaster is not fitted yet. Call `fit` with appropriate "
         "arguments before using `get_feature_importances()`.")
    )
    with pytest.raises(NotFittedError, match = err_msg):         
        forecaster.get_feature_importances()


def test_output_get_feature_importances_when_regressor_is_RandomForest():
    """
    Test output of get_feature_importances when regressor is RandomForestRegressor with lags=3
    and it is trained with series pandas DataFrame.
    """
    forecaster = ForecasterAutoregMultiSeries(
                     regressor = RandomForestRegressor(n_estimators=1, max_depth=2, random_state=123),
                     lags     = 3,
                     encoding = 'onehot'
                 )
    forecaster.fit(series=series)

    results = forecaster.get_feature_importances(sort_importance=False)
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', '1', '2'],
                   'importance': np.array([0.9446366782006932, 0.0, 0.05536332179930687, 0.0, 0.0])
               })

    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_regressor_is_RandomForest_with_exog():
    """
    Test output of get_feature_importances when regressor is RandomForestRegressor with lags=3
    and it is trained with series pandas DataFrame and a exogenous variable
    exog=pd.Series(np.arange(10, 20), name='exog').
    """
    forecaster = ForecasterAutoregMultiSeries(
                     regressor = RandomForestRegressor(n_estimators=1, max_depth=2, random_state=123),
                     lags     = 3,
                     encoding = 'onehot'
                 )
    forecaster.fit(series=series, exog=pd.Series(np.arange(10, 20), name='exog'))
    
    results = forecaster.get_feature_importances()
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', '1', '2', 'exog'],
                   'importance': np.array([0.73269896, 0., 0.21193772, 0., 0., 0.05536332])
               }).sort_values(by='importance', ascending=False)

    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_regressor_is_LinearRegression():
    """
    Test output of get_feature_importances when regressor is LinearRegression with lags=3
    and it is trained with series pandas DataFrame.
    """
    forecaster = ForecasterAutoregMultiSeries(
                     regressor = LinearRegression(),
                     lags      = 3,
                     encoding  = 'onehot'
                 )
    forecaster.fit(series=series)

    results = forecaster.get_feature_importances(sort_importance=False)
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', '1', '2'],
                   'importance': np.array([3.33333333e-01, 3.33333333e-01, 3.33333333e-01, 
                                           -1.48164367e-16, 1.48164367e-16])
               })

    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_regressor_is_LinearRegression_with_exog():
    """
    Test output of get_feature_importances when regressor is LinearRegression with lags=3
    and it is trained with series pandas DataFrame and a exogenous variable
    exog=pd.Series(np.arange(10, 15), name='exog').
    """
    series_2 = pd.DataFrame({'1': pd.Series(np.arange(10)), 
                             '2': pd.Series(np.arange(10))})

    forecaster = ForecasterAutoregMultiSeries(
                     regressor          = LinearRegression(),
                     lags               = 3,
                     encoding           = 'onehot',
                     transformer_series = None
                 )
    forecaster.fit(series=series_2, exog=pd.Series(np.arange(10, 20), name='exog'))

    results = forecaster.get_feature_importances(sort_importance=False)
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', '1', '2', 'exog'],
                   'importance': np.array([2.50000000e-01,  2.50000000e-01,  2.50000000e-01,  
                                           -2.97120907e-17,  2.97120907e-17, 2.50000000e-01])
               })

    pd.testing.assert_frame_equal(results, expected)


def test_output_and_UserWarning_get_feature_importances_when_regressor_no_attributes():
    """
    Test output of get_feature_importances when regressor is MLPRegressor with lags=3
    and it is trained with series pandas DataFrame. Since MLPRegressor hasn't attributes
    `feature_importances_` or `coef_, results = None and a UserWarning is issues.
    """
    series_2 = pd.DataFrame({'1': pd.Series(np.arange(5)), 
                             '2': pd.Series(np.arange(5))})
    forecaster = ForecasterAutoregMultiSeries(MLPRegressor(solver = 'lbfgs', max_iter= 50, random_state=123), lags=3)
    forecaster.fit(series=series_2)

    estimator = forecaster.regressor
    expected = None

    warn_msg = re.escape(
        (f"Impossible to access feature importances for regressor of type "
         f"{type(estimator)}. This method is only valid when the "
         f"regressor stores internally the feature importances in the "
         f"attribute `feature_importances_` or `coef_`.")
    )
    with pytest.warns(UserWarning, match = warn_msg):
        results = forecaster.get_feature_importances()
        assert results is expected


def test_output_get_feature_importances_when_pipeline_LinearRegression():
    """
    Test output of get_feature_importances when regressor is pipeline,
    (StandardScaler() + LinearRegression with lags=3),
    it is trained with series pandas DataFrame.
    """
    forecaster = ForecasterAutoregMultiSeries(
                     regressor          = make_pipeline(StandardScaler(), LinearRegression()),
                     lags               = 3,
                     encoding           = 'onehot',
                     transformer_series = None
                 )
    forecaster.fit(series=series)

    results = forecaster.get_feature_importances(sort_importance=False)
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', '1', '2'],
                   'importance': np.array([6.66666667e-01, 6.66666667e-01, 6.66666667e-01, 
                                           -7.40821837e-17, 7.40821837e-17])
               })

    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_pipeline_RandomForestRegressor():
    """
    Test output of get_feature_importances when regressor is pipeline,
    (StandardScaler() + RandomForestRegressor with lags=3),
    it is trained with series pandas DataFrame.
    """
    forecaster = ForecasterAutoregMultiSeries(
                     regressor = make_pipeline(
                                     StandardScaler(), 
                                     RandomForestRegressor(n_estimators=1, max_depth=2, random_state=123)
                                 ),
                     lags               = 3,
                     encoding           = "onehot",
                     transformer_series = None,
                 )
    forecaster.fit(series=series)

    results = forecaster.get_feature_importances(sort_importance=False)
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', '1', '2'],
                   'importance': np.array([0.9446366782006932, 0.0, 0.05536332179930687, 0.0, 0.0])
               })

    pd.testing.assert_frame_equal(results, expected)


# Unit test grid_search_sarimax
# ==============================================================================
import re
import pytest
import platform
import numpy as np
import pandas as pd
from skforecast.Sarimax import Sarimax
from skforecast.ForecasterSarimax import ForecasterSarimax
from skforecast.model_selection_sarimax import grid_search_sarimax

from tqdm import tqdm
from functools import partialmethod
tqdm.__init__ = partialmethod(tqdm.__init__, disable=True) # hide progress bar

# Fixtures
from ...ForecasterSarimax.tests.fixtures_ForecasterSarimax import y_datetime


def test_output_grid_search_sarimax_sarimax_with_mocked():
    """
    Test output of grid_search_sarimax in ForecasterSarimax with mocked
    (mocked done in Skforecast v0.7.0).
    """
    forecaster = ForecasterSarimax(
                     regressor = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    param_grid = [{'order' : [(2, 2, 0), (3, 2, 0)],
                   'trend': [None, 'c']}]

    results = grid_search_sarimax(
                  forecaster         = forecaster,
                  y                  = y_datetime,
                  param_grid         = param_grid,
                  steps              = 3,
                  refit              = False,
                  metric             = 'mean_absolute_error',
                  initial_train_size = len(y_datetime)-12,
                  fixed_train_size   = False,
                  return_best        = False,
                  verbose            = False
              )
    
    expected_results = pd.DataFrame(
        data  = {'params'             : [{'order': (3, 2, 0), 'trend': None},
                                         {'order': (3, 2, 0), 'trend': 'c'}, 
                                         {'order': (2, 2, 0), 'trend': 'c'}, 
                                         {'order': (2, 2, 0), 'trend': None}],
                 'mean_absolute_error': np.array([0.15357204, 0.1548934 , 0.19852912, 0.19853423]),
                 'order'              : [(3, 2, 0), (3, 2, 0), (2, 2, 0), (2, 2, 0)],
                 'trend'              : [None, 'c', 'c', None]},
        index = pd.Index(np.array([2, 3, 1, 0]), dtype='int64')
    )

    pd.testing.assert_frame_equal(results, expected_results, atol=0.0001)


# Unit test get_feature_importances ForecasterAutoreg
# ==============================================================================
import re
import pytest
from pytest import approx
import numpy as np
import pandas as pd
from skforecast.ForecasterAutoreg import ForecasterAutoreg
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LinearRegression
from sklearn.neural_network import MLPRegressor
from sklearn.ensemble import RandomForestRegressor


def test_NotFittedError_is_raised_when_forecaster_is_not_fitted():
    """
    Test NotFittedError is raised when calling get_feature_importances() and 
    forecaster is not fitted.
    """
    forecaster = ForecasterAutoreg(
                     regressor = LinearRegression(),
                     lags = 3,
                 )

    err_msg = re.escape(
                ("This forecaster is not fitted yet. Call `fit` with appropriate "
                 "arguments before using `get_feature_importances()`.")
              )
    with pytest.raises(NotFittedError, match = err_msg):         
        forecaster.get_feature_importances()


def test_output_get_feature_importances_when_regressor_is_RandomForest():
    """
    Test output of get_feature_importances when regressor is RandomForestRegressor with lags=3
    and it is trained with y=pd.Series(np.arange(10)).
    """
    forecaster = ForecasterAutoreg(RandomForestRegressor(n_estimators=1, max_depth=2, random_state=123), lags=3)
    forecaster.fit(y=pd.Series(np.arange(10)))
    
    results = forecaster.get_feature_importances()
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3'],
                   'importance': np.array([0.94766355, 0., 0.05233645])
               }).sort_values(by='importance', ascending=False)

    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_regressor_is_RandomForest_with_exog():
    """
    Test output of get_feature_importances when regressor is RandomForestRegressor with lags=3
    and it is trained with y=pd.Series(np.arange(10)) and a exogenous variable
    exog=pd.Series(np.arange(10, 20), name='exog').
    """
    forecaster = ForecasterAutoreg(RandomForestRegressor(n_estimators=1, max_depth=2, random_state=123), lags=3)
    forecaster.fit(y=pd.Series(np.arange(10)), exog=pd.Series(np.arange(10, 20), name='exog'))
    
    results = forecaster.get_feature_importances()
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', 'exog'],
                   'importance': np.array([0.94766355, 0.05233645, 0., 0.])
               })

    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_regressor_is_LinearRegression():
    """
    Test output of get_feature_importances when regressor is LinearRegression with lags=3
    and it is trained with y=pd.Series(np.arange(5)).
    """
    forecaster = ForecasterAutoreg(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(5)))

    results = forecaster.get_feature_importances(sort_importance=False)
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3'],
                   'importance': np.array([0.33333333, 0.33333333, 0.33333333])
               })

    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_regressor_is_LinearRegression_with_exog():
    """
    Test output of get_feature_importances when regressor is LinearRegression with lags=3
    and it is trained with y=pd.Series(np.arange(5)) and a exogenous variable
    exog=pd.Series(np.arange(10, 15), name='exog').
    """
    forecaster = ForecasterAutoreg(LinearRegression(), lags=3)
    forecaster.fit(y=pd.Series(np.arange(5)), exog=pd.Series(np.arange(10, 15), name='exog'))
    
    results = forecaster.get_feature_importances(sort_importance=False)
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3', 'exog'],
                   'importance': np.array([0.25, 0.25, 0.25, 0.25])
               })

    pd.testing.assert_frame_equal(results, expected)


def test_output_and_UserWarning_get_feature_importances_when_regressor_no_attributes():
    """
    Test output of get_feature_importances when regressor is MLPRegressor with lags=3
    and it is trained with y=pd.Series(np.arange(10)). Since MLPRegressor hasn't attributes
    `feature_importances_` or `coef_, results = None and a UserWarning is issues.
    """
    forecaster = ForecasterAutoreg(MLPRegressor(solver = 'lbfgs', max_iter= 50, random_state=123), lags=3)
    forecaster.fit(y=pd.Series(np.arange(10)))

    estimator = forecaster.regressor
    expected = None

    warn_msg = re.escape(
                (f"Impossible to access feature importances for regressor of type "
                 f"{type(estimator)}. This method is only valid when the "
                 f"regressor stores internally the feature importances in the "
                 f"attribute `feature_importances_` or `coef_`.")
            )
    with pytest.warns(UserWarning, match = warn_msg):
        results = forecaster.get_feature_importances()
        assert results is expected


def test_output_get_feature_importances_when_pipeline_LinearRegression():
    """
    Test output of get_feature_importances when regressor is pipeline,
    (StandardScaler() + LinearRegression with lags=3),
    it is trained with y=pd.Series(np.arange(5)).
    """
    forecaster = ForecasterAutoreg(
                     regressor = make_pipeline(StandardScaler(), LinearRegression()),
                     lags      = 3
                 )
    forecaster.fit(y=pd.Series(np.arange(5)))

    results = forecaster.get_feature_importances(sort_importance=False)
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3'],
                   'importance': np.array([0.166667, 0.166667, 0.166667])
               })

    pd.testing.assert_frame_equal(results, expected)


def test_output_get_feature_importances_when_pipeline_RandomForestRegressor():
    """
    Test output of get_feature_importances when regressor is pipeline,
    (StandardScaler() + RandomForestRegressor with lags=3),
    it is trained with y=pd.Series(np.arange(5)).
    """
    forecaster = ForecasterAutoreg(
                     regressor = make_pipeline(StandardScaler(), 
                                               RandomForestRegressor(n_estimators=1, max_depth=2, random_state=123)),
                     lags      = 3
                 )
    forecaster.fit(y=pd.Series(np.arange(10)))

    results = forecaster.get_feature_importances(sort_importance=False)
    expected = pd.DataFrame({
                   'feature': ['lag_1', 'lag_2', 'lag_3'],
                   'importance': np.array([0.94766355, 0., 0.05233645])
               })

    pd.testing.assert_frame_equal(results, expected)


# Unit test multivariate_time_series_corr
# ==============================================================================
import re
import pytest
import pandas as pd
from skforecast.utils.utils import multivariate_time_series_corr


def test_output_multivariate_time_series_corr():
    """
    Test the output of the correlation matrix
    """

    time_series = pd.Series(range(10), name='series_1')
    other = pd.DataFrame({
                'series_2': range(10),
                'series_3': [9, 2, 3, 8, 2, 6, 0, 8, 6, 1]
            })

    expected = pd.DataFrame(
                data = {
                    'series_2': [1., 1., 1., 1., 1.],
                    'series_3': [-0.21854656, -0.02836, -0.11964501, -0.45360921, -0.17251639],
                },
                index = pd.RangeIndex(start=0, stop=5, step=1)
            )
    expected.index.name = 'lag'

    results = multivariate_time_series_corr(
                  time_series = time_series,
                  other       = other,
                  lags        = 5
              )

    pd.testing.assert_frame_equal(results, expected)


def test_exception_multivariate_when_inputs_have_different_index():
    """
    Test that an exception is raised when `time_Series` and `other` have different index.
    """

    time_series = pd.Series(range(5), name='series_1')
    other = pd.DataFrame(
                {'series_2': range(5),
                'series_3': [9, 2, 3, 8, 2]},
                index = pd.RangeIndex(start=10, stop=15, step=1)
            )

    err_msg = re.escape("`time_series` and `other` must have the same index.")
    with pytest.raises(ValueError, match = err_msg):
        multivariate_time_series_corr(
            time_series = time_series,
            other       = other,
            lags        = 5
        )


def test_exception_multivariate_when_inputs_have_different_length():
    """
    Test that an exception is raised when `time_Series` and `other` have different lenth.
    """

    time_series = pd.Series(range(10), name='series_1')
    other = pd.DataFrame({
                'series_2': range(5),
                'series_3': [9, 2, 3, 8, 2]
            })

    err_msg = re.escape("`time_series` and `other` must have the same length.")
    with pytest.raises(ValueError, match = err_msg):
        multivariate_time_series_corr(
            time_series = time_series,
            other       = other,
            lags        = 5
        )
    
